package netproj.hosts;

import netproj.skeleton.Device;
import netproj.skeleton.Packet;

/**
 * A base class for all hosts (endpoints in the network)
 * 
 * You should extend this class if you want to implement a host. When you want to send a packet
 * to the network, call the {@code sendPacket(Packet)} method.
 * 
 * This class keeps track of a default gateway (which is initialized to 0) and all packets are sent
 * to this default gateway.
 * 
 * For more technical details, refer to {@link Device}'s documentation.
 * 
 * @author emil
 */
public abstract class Host extends Device {
	private int defaultGateway = 0;
	
	/**
	 * The default gateway is the gate through which packets are sent to the network.
	 * 
	 * By default, it is 0.
	 */
	public int getDefaultGateway() {
		return defaultGateway;
	}

	/**
	 * Sets the default gateway.
	 * 
	 * The default gateway is the gate through which packets are sent to the network. By default,
	 * it is 0.
	 * @param defaultGateway
	 */
	public void setDefaultGateway(int defaultGateway) {
		this.defaultGateway = defaultGateway;
	}

	/**
	 * Creates a basic host, an endpoint in the network, with default gateway set to 0.
	 */
	public Host(int inputBuffSize, int outputBuffSize, int address) {
		super(inputBuffSize, outputBuffSize, address);
	}
	
	/**
	 * Sends a packet to the default gateway (which is 0 by default).
	 */
	protected final void sendPacket(Packet packet) {
		sendPacket(packet, defaultGateway);
	}
}
