package netproj.hosts.tcp;

import java.util.LinkedList;
import java.util.Queue;
import java.util.logging.Logger;
import netproj.hosts.Host;
import netproj.skeleton.Packet;

/**
 * Base class for TCP Hosts.
 * 
 * Supports sending files, handles timeouts and ACKs for you. You just need to implement the
 * {@code onAck()} and {@code onTimeout()} methods to update your window size according to whatever
 * algorithm you are using.
 * 
 * This class can also be used as a standalone ultra dumb TCP host.
 * 
 * @author emil
 */
public class TcpHost extends Host {
	private static int PACKET_SIZE = 1024 * 8;
	
	protected double windowSize = 0;
	private long timeout = 2000;
	private long lastId = 0;
	
	private ResenderThread resenderThread;
	private Queue<TcpPacket> outstandingPackets;
	
	private static final Logger logger = Logger.getLogger(TcpHost.class.getCanonicalName());
	
	// Roundtrip stats
	protected double rtt = -1, rttStdev = -1;
	
	public TcpHost(int inputBuffSize, int outputBuffSize, int address) {
		super(inputBuffSize, outputBuffSize, address);
		
		outstandingPackets = new LinkedList<TcpPacket>();
		resenderThread = new ResenderThread();
		resenderThread.start();
	}
	
	/**
	 * Sends a file through the network.
	 * 
	 * Starts a new session and sends {@code size} bits to {@code dest}.
	 */
	public void sendFile(int dest, int size) {
		// Reset stats for a new session
		timeout = 2000;
		windowSize = 1;
		for (int i = 0; i < Math.ceil(size / (double) PACKET_SIZE); ) {
			synchronized (outstandingPackets) {
				if (outstandingPackets.size() < windowSize) {
					TcpPacket packet = new TcpPacket(getAddress(), dest, PACKET_SIZE, lastId++);
					i++;
					sendPacket(packet);
					outstandingPackets.add(packet);
					outstandingPackets.notifyAll();
				} else {
					// Wait until an ACK is received
					try {
						outstandingPackets.wait();
					} catch (InterruptedException e) {
						// This should not happen!
						assert(false);
					}
				}
			}
		}
		synchronized (outstandingPackets) {
			while (!outstandingPackets.isEmpty()) {
				try {
					outstandingPackets.wait();
				} catch (InterruptedException e) {

				}
			}
		}
		windowSize = 0;
	}
	
	public double getRtt() {
		return rtt;
	}
	
	public double getWindowSize() {
		return windowSize;
	}
	
	/**
	 * Called whenever a packet times out.
	 * 
	 * Override this method to update your window size according to your algorithm.
	 */
	protected void onTimeout() {
		
	}
	
	/**
	 * Called whenever an ACK is received.
	 * 
	 * Override this method to update your window size according to your algorithm.
	 * 
	 * Note that you may receive multiple ACKs for the same packet.
	 */
	protected void onAck(TcpPacket packet) {
		
	}
	
	@Override
	public void processPacket(Packet packet) {
		if (packet.getDestAddress() == getAddress() && packet instanceof TcpPacket) {
			TcpPacket tcpPacket = (TcpPacket) packet;
			if (tcpPacket.isAck()) {
				// Update roundtrip stats
				if (rtt < -0.5) {
					// This is the first reply
					rtt = rttStdev = tcpPacket.getRtt();
				} else {
					double t = tcpPacket.getRtt();
					rtt = 0.9 * rtt + 0.1 * t;
					rttStdev = 0.9 * rttStdev + 0.1 * Math.abs(t - rtt);
				}
				timeout = (long) Math.ceil(rtt + 4 * rttStdev);

				synchronized (outstandingPackets) {
					for (TcpPacket p : outstandingPackets) {
						if (p.getId() == tcpPacket.getId()) {
							outstandingPackets.remove(p);
							outstandingPackets.notifyAll();
							break;
						}
					}
				}
				onAck(tcpPacket);
			} else {
				sendPacket(tcpPacket.acknowledge());
			}
		}
	}
	
	@Override
	public void handleCommand(String command) {
		if (command.equals("help")) {
			super.handleCommand("help");
			System.out.println("send <address> <size> - sends <size> bits to <address>");
			System.out.println("asend <address> <size> - sends <size> bits to <address> asynchronously");
		} else if (command.startsWith("send")) {
			int address = Integer.decode(command.split(" ")[1]);
			int size = Integer.decode(command.split(" ")[2]);
			long time = System.currentTimeMillis();
			System.out.println("Transfer started.");
			sendFile(address, size);
			time = System.currentTimeMillis() - time;
			System.out.println("Transfer completed in " + time + "ms. (bps: " + (size * 1000L / time) + ")");
		} else if (command.startsWith("asend")) {
			int address = Integer.decode(command.split(" ")[1]);
			int size = Integer.decode(command.split(" ")[2]);
			new SenderThread(address, size).start();
			System.out.println("Transfer started.");
		} else {
			super.handleCommand(command);
		}
	}
	
	class SenderThread extends Thread {
		private int address;
		private int size;
		
		public SenderThread(int address, int size) {
			this.address = address;
			this.size = size;
		}
		@Override
		public void run() {
			long time = System.currentTimeMillis();
			sendFile(address, size);
			time = System.currentTimeMillis() - time;
			logger.fine("Transfer from " + Integer.toHexString(this.address) + " to " +
					Integer.toHexString(address) + " of " + size + " bits completed in " +
					time + "ms. (bps: " + (size * 1000L / time) + ")");
		}
	}
	class ResenderThread extends Thread {
		@Override
		public void run() {
			while (!isInterrupted()) {
				synchronized (outstandingPackets) {
					try {
						if (outstandingPackets.isEmpty()) {
							outstandingPackets.wait();
							continue;
						}
						if (outstandingPackets.peek().getRtt() < timeout) {
							outstandingPackets.wait(timeout - outstandingPackets.peek().getRtt());
							continue;
						}
						TcpPacket packet = outstandingPackets.poll();
						packet.reset();
						sendPacket(packet);
						outstandingPackets.add(packet);
						onTimeout();
					} catch (InterruptedException e) {
						break;
					}
				}
			}
		}
	}
}
