package netproj.routers;

import java.util.HashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import netproj.skeleton.Packet;

/**
 * A router that implements the distance-vector based Bellman Ford routing algorithm
 * to dynamically generate a routing table.
 * 
 * The router does not need to know any information about the network topology to
 * correctly run and build its routing table; to achieve this, the router has two
 * special types of packets, {@code BFDiscoverPacket} and {@code BFWeightPacket},
 * that are used to explore the network and relay routing information to neighboring
 * routers.  It is assumed that all routers in the network implement Bellman Ford
 * routing, since non-Bellman Ford devices are typically not configured to handle
 * the special packet types.
 * 
 * Since no guarantees can be made about the connectivity and reliability of
 * networks in general, this router is not guaranteed to converge to a state in
 * which the network topology is fully known.  However, in practice, the frequency
 * with which exploration packets are sent, combined with the relatively small size
 * of the exploration packets, causes the routers to converge to a full understanding
 * of the network topology in a relatively short timespan (typically within 10 seconds
 * for networks with around 10 routers connected by 10Mbps links, for example).
 * 
 * 
 * @author nathan
 *
 */
public class BFRouter extends Router {
	private static final Logger logger = Logger.getLogger(BFRouter.class.getCanonicalName());

	// Maps router addresses of routers that are only one hop away to
	// (time it takes to make one hop, gate at which router resides) pairs.
	private HashMap<Integer, TimeGatePair> oneHop;
	private HashMap<Integer, TimeGatePair> timeEstimates;
	
	private OneHopPinger pinger;
	
	public BFRouter(int inputBuffSize, int outputBuffSize, int address) {
		super(inputBuffSize, outputBuffSize, address);
		oneHop = new HashMap<Integer, TimeGatePair>();
		timeEstimates = new HashMap<Integer, TimeGatePair>();
		timeEstimates.put(address, new TimeGatePair(0, -1)); // A router can always get to itself in 0 time
		pinger = new OneHopPinger();
	}
	
	public void start() {
		pinger.start();
		super.start();
	}
	
	public void processPacket(Packet msg) {
		if (msg instanceof BFWeightPacket) {
			// this is a packet of updated weights -- we only care if it's coming from one hop away
			BFWeightPacket wmsg = (BFWeightPacket)msg;
			TimeGatePair msgCost = oneHop.get(wmsg.getSourceAddress());
			if (msgCost != null) {
				// we have changes to consider, so prepare a packet to send
				BFWeightPacket newBroadcast = new BFWeightPacket(getAddress(), -1, 32*8, 1);
				for (WeightUpdate u : wmsg.getWeights()) {
					TimeGatePair currCost = timeEstimates.get(u.getDestAddress());
					int newWeight = msgCost.getTime() + u.getWeight();
					if (currCost == null) {
						// this is an all-new entry
						
						// create time estimate
						timeEstimates.put(u.getDestAddress(),
								new TimeGatePair(newWeight, msgCost.getGate()));
						
						// create routing table entry
						addRoutingTableEntry(u.getDestAddress(), 24, msgCost.getGate());
						
						// prepare the update to be broadcast
						newBroadcast.addWeight(u.getDestAddress(), newWeight);
					} else if (currCost.getGate() == msgCost.getGate()) {
						newWeight = (int)(.75 * currCost.getTime() + .25 * newWeight);
						// update time estimate
						timeEstimates.put(u.getDestAddress(),
								new TimeGatePair(newWeight, msgCost.getGate()));
						
						// prepare the update to be broadcast
						if (Math.abs(currCost.getTime() - newWeight) > 2)
							newBroadcast.addWeight(u.getDestAddress(), newWeight);
					} else if (newWeight < currCost.getTime()) {
						// update time estimate
						timeEstimates.put(u.getDestAddress(),
								new TimeGatePair(newWeight, msgCost.getGate()));
						
						// update routing table entry
						addRoutingTableEntry(u.getDestAddress(), 24, msgCost.getGate());
						
						// prepare the update to be broadcast
						newBroadcast.addWeight(u.getDestAddress(), newWeight);
					}
				}
				if (newBroadcast.getWeights().size() > 0) {
					// broadcast changes
					broadcastPacket(newBroadcast);
				}
			}
		} else if (msg instanceof BFDiscoverPacket) {
			// this is a packet for gathering one-hop information
			BFDiscoverPacket pmsg = (BFDiscoverPacket)msg;

			if (pmsg.getDestAddress() == -1) {
				// we're receiving a ping from one hop away, measure the time it took
				// and see if we can get it back to the sender
				int elapsed = (int)(System.currentTimeMillis() - pmsg.getTimestamp());

				BFDiscoverPacket broadcastReply = new BFDiscoverPacket(getAddress(),
						pmsg.getSourceAddress(), 32, 1, elapsed);
				broadcastReply.setSendGate(pmsg.getSendGate());
				BFDiscoverPacket routedReply = new BFDiscoverPacket(getAddress(),
						pmsg.getSourceAddress(), 32, 16, elapsed);
				routedReply.setSendGate(pmsg.getSendGate());

				broadcastReply.recordTimestamp();
				routedReply.recordTimestamp();
				broadcastPacket(broadcastReply);
				super.processPacket(routedReply);
			} else if (pmsg.getDestAddress() == getAddress()) {
				oneHop.put(pmsg.getSourceAddress(),
						new TimeGatePair(pmsg.getElapsedTime(), pmsg.getSendGate()));
			} else {
				// treat this like a normal packet if we aren't involved
				super.processPacket(msg);
			}
		} else {
			super.processPacket(msg);
		}
	}
	
	protected class TimeGatePair {
		private int time;
		private int gate;
		
		public TimeGatePair(int t, int g) {
			time = t;
			gate = g;
		}
		
		public int getTime() {
			return time;
		}
		public int getGate() {
			return gate;
		}
		
		public void setTime(int t) {
			time = t;
		}
		public void setGate(int g) {
			gate = g;
		}
	}
	
	private class OneHopPinger extends Thread {
		/**
		 * Periodically pings neighbors to keep one-hop data accurate
		 */
		@Override
		public void run() {
			while (!this.isInterrupted()) {
				for (int gate = 0; gate < getLinks().size(); gate++) {
					BFDiscoverPacket ping = new BFDiscoverPacket(getAddress(), -1, 32*8, 1);
					BFWeightPacket selfweight = new BFWeightPacket(getAddress(), -1, 32*8, 1);
					selfweight.addWeight(getAddress(), 0);
					ping.setSendGate(gate);
					
					ping.recordTimestamp();
					sendPacket(ping, gate);
					sendPacket(selfweight, gate);
				}
				
				try {
					synchronized (this) {
						wait(1000 + (int)(3000 * Math.random()));
					}
				} catch (InterruptedException e) {
					logger.log(Level.WARNING, "Pinger for router " + getAddress() + " was "
							   + "interrupted.", e);
					break;
				}
			}
		}
	}
}
