package netproj.routers;

import java.util.ArrayList;

import netproj.skeleton.Packet;

/**
 * A type of packet used specifically by {@code BFRouter} devices to learn
 * the topology of the network.  This type of packet relies on successful
 * transmission of {@code BFDiscoverPacket}s prior to becoming useful, since
 * the discover packets are responsible for associating nearby routers with
 * gates/links for each router.  Once the discover packets successfully inform
 * the router of which routers are only one hop away, receiving packets from
 * an adjacent router enables the router to use the information to compute
 * distance approximations to all devices listed in the packet.
 * 
 * 
 * @author nathan
 *
 */
public class BFWeightPacket extends Packet {
	private ArrayList<WeightUpdate> weights;
	
	public BFWeightPacket(int source, int dest, int size, int ttl) {
		super(source, dest, size, ttl);
		weights = new ArrayList<WeightUpdate>();
	}
	
	protected BFWeightPacket(BFWeightPacket p) {
		super(p);
		weights = new ArrayList<WeightUpdate>();
		for (WeightUpdate u : p.getWeights()) {
			weights.add(u);
		}
	}
	
	public Packet copy() {
		return new BFWeightPacket(this);
	}
	
	public ArrayList<WeightUpdate> getWeights() {
		return weights;
	}
	
	public void addWeight(int dest, int weight) {
		weights.add(new WeightUpdate(dest, weight));
		setSizeInBits(getSizeInBits() + 8 * 8);
	}
}

class WeightUpdate {
	private int dest;
	private int weight;
	
	public WeightUpdate(int dest, int weight) {
		this.dest = dest;
		this.weight = weight;
	}
	
	public int getDestAddress() {
		return dest;
	}
	
	public int getWeight() {
		return weight;
	}
}