package netproj.skeleton;

/**
 * IP packet.
 * 
 * If you are making a subclass of Packet, YOU MUST OVERRIDE {@code copy()}.
 * 
 * @author emil, nathan
 */
public class Packet {
	/**
	 * The largest used packet identifier so far. 
	 */
	private static long lastPacketId = 0;
	public int getTtl() {
		return ttl;
	}

	public void setTtl(int ttl) {
		this.ttl = ttl;
	}

	private static final Object packetIdSyncObj = new Object();
	
	/**
	 * Unique identifier of the packet. To be generated by the constructor.
	 */
	private final long packetId;
	
	private int source;
	private int dest;
	private int size;
	private int ttl;
	private long propDelayTime = 0;
	
	/**
	 * Creates a new IP packet.
	 * 
	 * @param source	IPv4 address of the source device.
	 * @param dest		IPv4 address of the destination device.
	 * @param size		Size of this packet in bits.
	 * @param ttl		Number of hops this packet can make before it is dropped.
	 */
	public Packet(int source, int dest, int size, int ttl) {
		this.source = source;
		this.dest = dest;
		this.size = size;
		this.ttl = ttl;
		
		synchronized (packetIdSyncObj) {
			lastPacketId++;
			packetId = lastPacketId;
		}
	}
	
	/**
	 * Creates a new IP packet.
	 * 
	 * @param source	IPv4 address of the source device.
	 * @param dest		IPv4 address of the destination device.
	 * @param size		Size of this packet in bits.
	 */
	public Packet(int source, int dest, int size) {
		this(source, dest, size, 64);
	}
	
	/**
	 * Copy constructor.
	 * 
	 * If you are making a subclass of Packet, YOU MUST OVERRIDE THIS METHOD.
	 */
	protected Packet(Packet p) {
		source = p.source;
		dest = p.dest;
		size = p.size;
		ttl = p.ttl;
		packetId = p.packetId;
		propDelayTime = p.propDelayTime;
	}
	
	/**
	 * Makes a copy of this packet.
	 * 
	 * This method is called when packets are sent on the network.
	 * 
	 * If you are making a subclass of Packet, YOU MUST OVERRIDE THIS METHOD!
	 * @return  a copy of the Packet
	 */
	public Packet copy() {
		return new Packet(this);
	}
	
	/**
	 * Adds another hop (decreases ttl by one).
	 * 
	 * @return	{@code true} if ttl >= 0 after the hop (the packet can hop) or {@code false} if the
	 * 			packet must be dropped.
	 */
	public boolean hop() {
		ttl--;
		if (ttl < 0) {
			return false;
		}
		return true;
	}
	
	public long getPropDelayTime() {
		return propDelayTime;
	}
	
	/**
	 * Unique packet ID.
	 */
	public long getPacketId() {
		return packetId;
	}
	
	/**
	 * Size of the packet in the network (in bits).
	 */
	public int getSizeInBits() {
		return size;
	}
	
	/**
	 * IPv4 address of the origin of this packet.
	 */
	public int getSourceAddress() {
		return source;
	}
	
	/**
	 * IPv4 address of the destination device.
	 */
	public int getDestAddress() {
		return dest;
	}
	
	public void setPropDelayTime(long delay) {
		propDelayTime = System.currentTimeMillis() + delay;
	}
	
	/**
	 * Sets the IPv4 address of the origin of this packet.
	 */
	public void setSourceAddress(int address) {
		source = address;
	}
	
	/**
	 * Sets the size (in bits) of the packet
	 */
	protected void setSizeInBits(int size) {
		this.size = size;
	}
	
	/**
	 * Sets the IPv4 address of the destination device.
	 */
	public void setDestAddress(int address) {
		dest = address;
	}
	
	public String toString() {
		return "(" + Integer.toString(source, 16) + ", " + Integer.toString(dest, 16) + ", " 
		       + size + ")";
	}
}
